
const NOT_FOUND_ERROR = "<html><head><title>404 Not Found</title>\n" + 
  "</head><body><h1>Not Found</h1>\n" + 
  "The requested URL %URL% was not found\n" +
  "<hr>\n<address>Zipedia: offline wikipedia browser</address>\n" + 
  "</body></html>";

const WIKI_SEARCH_FORM = '<form name="searchform" id="searchform" method="get" action="wikipedia://wikitools/search">\n      <input accesskey="s" id="searchInput" name="id" type="text" />\n <input value="Search" type="submit" class="searchButton" />\n      </form>\n';

const HTML_HEAD = '<html><head>\n		<meta http-equiv="Content-Type" content="text/html; charset=utf-8" />\n		<title>%TITLE%</title>\n</head>\n<body class="ns-0 ltr">\n<div id="globalWrapper"><div id="search" class="search">' + WIKI_SEARCH_FORM + '</div>\n<div id="column-content">\n<div id="content">\n<a name="top" id="top"></a><h1>%TITLE%</h1><hr/>\n';

const HTML_TAIL = '</div></div></div><center>[<a href="wikipedia://raw/%TITLE%">Source</a>] [<a href="wikipedia://wiki/%TALKNS%:%TITLE%">Talk</a>]</body>\n</html>';


function Wiki(wikiData) {    
    
    trace("Zipedia: Wiki()");

    this.handleWiki = handleWiki;
    this.handleRaw = handleRaw;
    this.handleSearch = handleSearch;
    this.redirect = redirect;

    /////////////////////// implementation ///////////////////
    

    function handleWiki(key, uri) {
	trace("wikipedia: handleWiki(" + key + ", " + uri.spec + ")");
	
	var text;

	try {
	    
	    var special = _isSpecialPage(key);
	    
	    if (special) {
		trace("wikipedia: handleWiki: calling _handleSepcial(" + special + ")");
		text = _handleSpecial(special);
	    } else {	  
	  
		text = _getEntry(key);

		trace("wikipedia: handleWiki: checking for redirections for key=" + key);
		
		var r = text.match(/^#REDIRECT\s*\[\[([^\]]+)\]\]/i);
		
		if (r && r[1] != key /* prevent loop */) {
		    trace("wikipedia: #REDIRECT: " + key + "=>" + r[1]);
		    
		    return redirect("wiki", r[1]);
		    
		    
		    
		    //return handleWiki(r[1], uri);
		} 
	    }


	    text = _format(text, uri.spec);

	} catch (e) {
	    trace("wikipedia: handleWiki: e=" + e);
	    return "" + e;
	}
	

	return postProcess(text, key);
    }

    function handleRaw(key) {
	var text;

	try {
	    text = _getEntry(key);
	} catch (e) {
	    text = "not found";
	}

	return text;
    }

    function handleSearch(key, query) {

	key = key ? key : query["id"];
    
	var pattern = query["id"] ? query["id"] : key;

	var namespace = query["ns"] ? query["ns"] : "";

	var text = "";


	var list = [];

	var it = wikiData.findArticle(namespace, pattern);
	  
	_getResults(it, list);
    
	var title = "Search Results for " + key;

	switch (list.length) {
	case 1:
	    key = list[0];
      
	    if (key[0] == ":") {
		key = key.substr(1);
	    }
      
	    return redirect("wiki", key);
	case 0:
	    text = "Nothing found";
	    break;
	default:
      
	    text = "<h2>" + list.length + " items:</h2>\n<ol>\n";
    
	    for (var i = 0; i < list.length; ++i) {
		text += '<li><a href="wikipedia://wiki/' + encodeURI(list[i]) + '">' + Utf8.encode(list[i]) + '</a>\n';
	    }
	    text += "</ol>"
		break;
	}

	return postProcess(text, title);
    }


    ///////////////////////////// private //////////////////////


    function _format(txt, url) {
	return wikiData.format(txt, url);
    }

    function postProcess(text, key) {

	var ns = "";

	if (!_isSpecialPage(key)) { 
	    ns = _getNamespace(key);
    
	    var talk;

	    if (ns != "") {
		talk = ns.substr(0, ns.length - 1) + " talk";
	    } else {
		talk = "Talk";
	    }
	}

	return HTML_HEAD.replace(/%TITLE%/g, key).replace(/%NS%/g, ns) + text + HTML_TAIL.replace(/%TITLE%/g, key).replace(/%TALKNS%/g, talk);
    }

    function redirect(proto, path, query) {
	
	if (query) {
	    path += "?dummy=";
      
	    for (var n in query) {
		path += "&" + n + "=" + query[n];
	    }
	}
	return '<html><head><meta http-equiv="Content-Type" content="text/html; charset=utf-8" /><meta HTTP-EQUIV="Refresh" CONTENT="0; URL=wikipedia://' + proto + '/' + encodeURI(path) + '"/></head></html>';
    }

    function _getResults(it, list) {
	var count = 0;
    
	while (it.hasMore() && count++ < 200) {
	    var n = it.getNext();

	    var ns = _getNamespace(n);

	    if (ns == ":") {
		n = n.substr(1);
	    }

	    list.push(n);
	}
    }
    function _getNamespace(key) {
	var list = key.split(':');

	if (list.length > 1) {
	    var ns = list[0];
	    var it = wikiData.getNamespaces();

	    while (it.hasMore()) {
		if (ns == it.getNext()) {
		    return ns;
		}
	    }
	}
	return "";
    }

    function _getEntry(key) {

	var ns = _getNamespace(key);

	var title = ns ? key.substr(ns.length + 1) : key;

	var txt =  wikiData.getArticle(ns, title, true);
	
	return txt;
    }

    function _isSpecialPage(key) {
	switch (key.toLowerCase()) {
	case "main page":
	case "main_page":
	case "mainpage":
	return "Main Page";
	}
    
	var r;

	if (null != (r = key.match(/^examples\/(.*)/i))) {
	    return "Examples/" + r[1];
	} else {
	    return null;
	}
    }

    function _handleSpecial(key) {
	var r;

	if (key == "Main Page") {
	    return getContents("chrome://zipedia/content/etc/MainPage.wp");
	} else if (null != (r = key.match("Examples/(.*)"))) {
	    if (!r[1]) {
		// list all examples under chrome://zipedia/content/etc/examples

		var text = "";

		var dir = buildPath(getExtensionPath("zipedia", true), ["chrome", "content", "zipedia", "etc", "examples"]);
	
		var i = dir.directoryEntries;

		while (i.hasMoreElements()) {
		    var file = i.getNext().QueryInterface(Components.interfaces.nsILocalFile);

		    if (file.isFile() && file.isReadable()) {
			var parts = file.leafName.split(".");

			if (parts.length == 2 && parts[1] == "wp") {
			    text += "* [[Examples/" + parts[0] + "]]\n";
			}
		    }
		}

		return text;
	    }

	    return getContents("chrome://zipedia/content/etc/examples/" + r[1] + ".wp");
	}      

	throw "HDIGH!";
    }
}