
function WikiDataSqlite() {

  this.wrappedJSObject = this;
}

WikiDataSqlite.prototype = {
  dbConn: null,
  dbStmntPage: null,
  dbStmntGlob: null,
  dbStmntLike: null,
  dbStmntNamespaces: null,
  dataFiles: {},


  loading: false,

  close: function() {

	if (this.loaded) {
	    [this.dbStmntNamespaces, this.dbStmntLike, this.dbStmntGlob, this.dbStmntPage].forEach(function(stmnt) {
		    if (stmnt && "finalize" in stmnt) {
			stmnt.finalize();
		    }
		});
	
	    if (this.dbConn && "close" in this.dbConn) {
		this.dbConn.close();
		this.dbConn = null;
	    }
	
	    for (var i in this.dataFiles) {
		this.dataFiles[i].close();
	    }
	
	    this.dataFiles = null;

	    this.loaded = false;
	}
  },    

  load: function() {
	if (!this.loaded) {

		
	    this.loading = true;


	    try {
		var dir;

		this.dirpath = YAGI_getPref("zipedia.", "zip.path");
		
		if (!this.dirpath) {
		    this.dirpath = buildPath(getExtensionPath("zipedia", true), ["zipdata"]).path;
		}

		dir = buildPath(this.dirpath);
	
		if (!dir.isDirectory()) {
		    throw this.dirpath + ": given zipedia dir path is not a directory";
		}

		this.dataFiles = {};

		{ // the connection
		    var storageService = Components.classes["@mozilla.org/storage/service;1"] 
			.getService(Components.interfaces.mozIStorageService);
	  

		    var sqlpath = buildPath(dir, ["zipedia.sq"]);
		    
		    if (!sqlpath.exists()) {
			throw "no zipedia.sq under " + sqlpath.path + " found";
		    }
      

		    this.dbConn = storageService.openDatabase(sqlpath);
	  
		    this.dbStmntPage = this.dbConn.createStatement("SELECT title, name, file, pos, size FROM page JOIN namespace ON page.ns_id = namespace.ns_id WHERE name = ?1 AND title = ?2");
	  
		    this.dbStmntLike = this.dbConn.createStatement("SELECT title, name, file, pos, size FROM page JOIN namespace ON page.ns_id = namespace.ns_id WHERE name LIKE ?1 AND title LIKE ?2");
	  
		    this.dbStmntGlob = this.dbConn.createStatement("SELECT title, name, file, pos, size FROM page JOIN namespace ON page.ns_id = namespace.ns_id WHERE name GLOB ?1 AND title GLOB ?2");
	  
		    this.dbStmntNamespaces = this.dbConn.createStatement("SELECT name FROM namespace ORDER BY name");
		}
	    } finally {
		this.loading = false;
		//this.synchLock.writeUnlock();
	    }

	    this.loaded = true;
	}
    },

  _findEntries: function(aPattern, mode) {
	trace("zipedia: _findEntries(" + aPattern + ", " + mode + ")");

	var stmnt;

	switch (mode) {
	case "like":
	    stmnt = this.dbStmntLike;
	    break;
	case "glob":
	    stmnt = this.dbStmntGlobe;
	    break;
	default:
	    stmnt = this.dbStmntPage;
	    break;
	}

	var args = aPattern.split(":");

	var ns = args.shift();
	
	var title = args.join(":");

	trace("zipedia: [ns,title] = [" + ns + ", " + title + "]");

	stmnt.bindUTF8StringParameter(0, ns);
	stmnt.bindUTF8StringParameter(1, title);


	return stmnt;
	
  },
  findEntries: function(aPattern) {

	aPattern = aPattern.replace(/\*/g,"%");
	
	var stmnt = this._findEntries(aPattern, "like");
	
	try {

	    var names = [];
	    var count = 0;

	    while (stmnt.executeStep() && count++ < 100) {
		var ns = stmnt.getUTF8String(1);
		var namespace = ns ? ns + ":" : "";
		var title = stmnt.getUTF8String(0);

		var entry = namespace + title;
		

		names.push(entry);
	    }


	    return {
		hasMore: function() {
		    return names.length;
		},
		    getNext: function() {
		    return names.shift();
		}
	    };
	} finally {
	    stmnt.reset();
	}	
  },
  _getIndex: function(key) {

	var stmnt = this._findEntries(key);
	
	try {
	    if (!stmnt.executeStep()) {
	    
		key = key.replace(/ /g,"_");
	    
		stmnt = this._findEntries(key, "like");

	    
		if (!stmnt.executeStep()) {
		    trace("entry/s for key " + key + " was not found");
		    return null;
		}
	    } 

	    var res = {
		file: "page-" + stmnt.getInt32(2) + ".dat",
		pos: stmnt.getInt32(3), 
		len: stmnt.getInt32(4)
	    };
	
	    return res;
	} finally {
	    stmnt.reset();
	    
	}
    },

  _openDataFile: function(name) {

	var file = buildPath(this.dirpath, [name]);

	if (!file.exists()) {
	    throw "data file " + file.path + " does not exist";
	}

	var fins = Components.classes["@mozilla.org/network/file-input-stream;1"].
	createInstance(Components.interfaces.nsIFileInputStream);

	fins.init(file, -1, -1, false);

	return fins;
    },
  _mkAsyncListener: function(output) {
    var listener = {
      output: output,
      onDataAvailable: function( request ,  context ,  inputStream ,  offset ,  count ) {
	this.output.writeFrom(inputStream, count);
      },
      onStartRequest: function (  request ,  context ) {

      },
      onStopRequest: function (  request ,  context ,  statusCode ) {
	this.output.close();
      }
    };
    
    return listener;
  },
  getEntry: function(namespace, key) {

	var index = this._getIndex(namespace, key);

	var text;
	
	if (null != index) {
	    
	    var ifstream = this._openDataFile(index.file);
	    
	    var ins = _getZstream(ifstream, index.pos, index.len);
	    
	    try {
		trace("zipedia: getEntry: text = readStream(ins) >>");
		text = readStream(ins);
		trace("zipedia: getEntry: text = readStream(ins) <<");
	    } finally {
		ifstream.close();
		ins.close();
	    }
	} else {
	    text = "Not found.";
	}

	return text;
    },
    getNamespaces: function() {

	var stmnt = this.dbStmntNamespaces;

	try {

	    var names = [];
	
	    while (stmnt.executeStep()) {
		names.push(stmnt.getUTF8String(0));
	    }

	    return {
		hasMore: function() {
		    return names.length;
		},
		getNext: function() {
		    return names.shift();
		}
	    };	    
	} finally {
	    stmnt.reset();	    
	}
    }
}


function _makeStreamConv(listener) {
  var streamConv = Components.classes["@mozilla.org/streamconv;1?from=deflate&to=uncompressed"].createInstance(Components.interfaces.nsIStreamConverter);
  streamConv.asyncConvertData("deflate", "uncompressed", listener, null);
  
  return streamConv;
}

function _makePipe() {
  var pipe = Components.classes["@mozilla.org/pipe;1"].
    createInstance(Components.interfaces.nsIPipe);
  
  pipe.init(false, false, 0, 0, null);
  
  return pipe;
}

function _getZstream(ifstream, off, len) {

  var ssl = Components.classes["@mozilla.org/network/sync-stream-listener;1"].
    createInstance(Components.interfaces.nsISyncStreamListener);

  var streamConv = _makeStreamConv(ssl);

  ifstream.QueryInterface(Components.interfaces.nsISeekableStream).seek(0, off);

  streamConv.onDataAvailable(null ,  null ,  ifstream ,  off ,  len );
  streamConv.onStopRequest (  null ,  null ,  0 );

  return ssl.inputStream;
}
