// Custom protocol related
const kSCHEME = "wikipedia";
const kPROTOCOL_CID = Components.ID("{C7715470-4AE3-4F14-B7E2-FB6974B5D25B}");
const kPROTOCOL_CONTRACTID = "@mozilla.org/network/protocol;1?name=" + kSCHEME;
const kPROTOCOL_NAME = "Embeded Wikipedia Protocol";


const kSTANDARDURL_CONTRACTID = "@mozilla.org/network/standard-url;1";
const kURLTYPE_STANDARD = 1;
const nsIComponentRegistrar = Components.interfaces.nsIComponentRegistrar;
const nsIFactory = Components.interfaces.nsIFactory;
const nsIProtocolHandler = Components.interfaces.nsIProtocolHandler;
const nsIStandardURL = Components.interfaces.nsIStandardURL;
const nsIURI = Components.interfaces.nsIURI;

const nsIThreadManager = Components.interfaces.nsIThreadManager;

const FF3 = !(!nsIThreadManager);

var tracingEnabled = false;

function makeZipediaChannel(key, type, uri, query) {

    var runInThread = YAGI_getPref("zipedia.backend.", "threaded");

    trace("zipedia: makeZipediaChannel(" + key + ", " + type + ", " + uri + ", " + query);


    var data;
    var contentType;
    var channel;
    var wikiData;

    try {

	wikiData = makeWikiDataInstance();

	if (runInThread) {
	    var nsIInputStreamChannel = Components.interfaces.nsIInputStreamChannel;

	    channel = Components.classes["@mozilla.org/network/input-stream-channel;1"].createInstance(nsIInputStreamChannel).QueryInterface(nsIInputStreamChannel);
	    var pipe = _makePipe();

	    //pipe.init(true, true, 0, 0, null);

	    channel.setURI(uri);
	    channel.contentStream = pipe.inputStream;

	    if (!FF3) {
		channel.contentType = type == "raw" ? "text/plain" : "text/html";
		channel.contentCharset = "utf8";
		channel.contentLength = 10000000000; //-1;
	    }




	    var runnable = {

		run: function() {		    

		    try {

			trace("Zipedia: runnable.run()");

			[data, contentType] = getData(wikiData, key, type, uri, query);
		
			trace("data: " + data.length + ", " + contentType);

			if (!FF3) {

			    channel.contentLength = data.length;
			
			    channel.contentType = contentType;
			}

			pipe.outputStream.write(data, data.length);

		    } catch (e) {
			[data, contentType] = ["Zipedia internal error: " + e, "text/plain"];

			if (!FF3) {
			    channel.contentType = contentType;
			}

			try {
			    pipe.outputStream.write(data, data.length);
			} catch (e) {
			    trace("runnable.run(): pipe.outputStream.write(data, data.length)" + e);
			}

		    } finally {
			try {
			    pipe.outputStream.close();
			} catch (e) {}
		    }
		}
	    };
  
	    getThread(runnable);
	    //runnable.run();

	} else {
	    [data, contentType] = getData(wikiData, key, type, uri, query);	    
	}
    } catch (e) {
	
	trace(e);
	
	[data, contentType] = ["" + e, "text/plain"];

	channel = null;
	
    }

    if (null == channel) {
	var ios = Components.classes["@mozilla.org/network/io-service;1"].getService(Components.interfaces.nsIIOService);
	channel = ios.newChannel("data:" + contentType + ";charset=utf-8," + data, "utf-8", uri);
    }

    return channel;
}


var ZipediaModule = {
  
  /* CID for this class */
  cid: kPROTOCOL_CID,

  /* Contract ID for this class */
  contractId: kPROTOCOL_CONTRACTID,

  registerSelf : function(compMgr, fileSpec, location, type) {
    compMgr = compMgr.QueryInterface(nsIComponentRegistrar);
    compMgr.registerFactoryLocation(
      kPROTOCOL_CID, 
      kPROTOCOL_NAME, 
      kPROTOCOL_CONTRACTID, 
      fileSpec, 
      location,
      type
    );
  },
  
  getClassObject : function(compMgr, cid, iid) {
    if (!cid.equals(kPROTOCOL_CID)) {
      throw Components.results.NS_ERROR_NO_INTERFACE;
    }
    if (!iid.equals(nsIFactory)) {
      throw Components.results.NS_ERROR_NOT_IMPLEMENTED;
    }
    return this.myFactory;
  },
  
  canUnload : function(compMgr) {
    return true;
  },
  
  myFactory : {
    createInstance : function(outer, iid) {
      if (outer != null) {
        throw Components.results.NS_ERROR_NO_AGGREGATION;
      }
                        
      return new ZipediaHandler().QueryInterface(iid);
    }
  }
};

function NSGetModule(compMgr, fileSpec) {
    return ZipediaModule;
}

/*----------------------------------------------------------------------
 * The Zipedia Handler
 *----------------------------------------------------------------------
 */

function ZipediaHandler() {
  trace("[ZipediaHandler.<init>]");
  
  this.wrappedJSObject = this;
  
  this._system_principal = null;
  
}

ZipediaHandler.prototype = {

    scheme: kSCHEME,
  
    defaultPort : -1,
  
    protocolFlags : nsIProtocolHandler.URI_STD,
  
  
    allowPort : function(port, scheme) {
	trace("[ZipediaHandler.allowPort]");
    
	return false;
    },
    newURI : function(spec, charset, baseURI) {
	trace("[ZipediaHandler.newURI] " + spec);
      
	var new_url = Components.classes[kSTANDARDURL_CONTRACTID].createInstance(nsIStandardURL);
	new_url.init(kURLTYPE_STANDARD, -1, spec, charset, null);    
    
	var new_uri = new_url.QueryInterface(nsIURI);
	return new_uri;
    },  
    _parseQuery: function (uri) {
    
	var query = {};

	if (-1 != uri.path.indexOf('?')) {

	    var query_list = uri.path.split('?').pop().split("&");
      
	    while (query_list.length) {
		var q = query_list.pop();
		var qv = q.split('=');
	
		query[qv[0]] = urldecode(qv[1]);
	    }
	}

	return query;
    },
    newChannel: function(origuri) {
	trace("[ZipediaHandler.newChannel] new channel requested for: " + origuri.spec);
	trace("[ZipediaHandler.newChannel(path)] new channel requested for: " + origuri.path);

	var uri = this.newURI(origuri.spec, "utf8"); // duplicate uri for path mannipulation

	//var uri = this.newURI(origuri.spec); // duplicate uri for path mannipulation

	var type = origuri.host;

	var key = uri.path.substr(1);

	key = key.split('?')[0]; // removes any query
	key = key.split('#')[0]; // removes any '#ref'

	//key = Utf8.encode(urldecode(key));
	key = urldecode(key);

	var query = this._parseQuery(origuri);

	return makeZipediaChannel(key, type, uri, query);
    },
    QueryInterface : function(iid) {
	trace("[ZipediaHandler.QueryInterface]");

	if (!iid.equals(Components.interfaces.nsIProtocolHandler) &&
	    !iid.equals(Components.interfaces.nsISupports)) {
      
	    trace("[ZipediaHandler.QueryInterface] error - NS_ERROR_NO_INTERFACE " + iid);
      
	    throw Components.results.NS_ERROR_NO_INTERFACE;
	}
	return this;
    }
};
const CHARTABLE = new Array('\x00', '\x01', '\x02', '\x03', '\x04', '\x05', '\x06', '\x07', '\x08', '\t', '\n', '\x0b', '\x0c', '\r', '\x0e', '\x0f', '\x10', '\x11', '\x12', '\x13', '\x14', '\x15', '\x16', '\x17', '\x18', '\x19', '\x1a', '\x1b', '\x1c', '\x1d', '\x1e', '\x1f', ' ', '!', '"', '#', '$', '%', '&', "'", '(', ')', '*', '+', ',', '-', '.', '/', '0', '1', '2', '3', '4', '5', '6', '7', '8', '9', ':', ';', '<', '=', '>', '?', '@', 'A', 'B', 'C', 'D', 'E', 'F', 'G', 'H', 'I', 'J', 'K', 'L', 'M', 'N', 'O', 'P', 'Q', 'R', 'S', 'T', 'U', 'V', 'W', 'X', 'Y', 'Z', '[', '\\', ']', '^', '_', '`', 'a', 'b', 'c', 'd', 'e', 'f', 'g', 'h', 'i', 'j', 'k', 'l', 'm', 'n', 'o', 'p', 'q', 'r', 's', 't', 'u', 'v', 'w', 'x', 'y', 'z', '{', '|', '}', '~', '\x7f', '\x80', '\x81', '\x82', '\x83', '\x84', '\x85', '\x86', '\x87', '\x88', '\x89', '\x8a', '\x8b', '\x8c', '\x8d', '\x8e', '\x8f', '\x90', '\x91', '\x92', '\x93', '\x94', '\x95', '\x96', '\x97', '\x98', '\x99', '\x9a', '\x9b', '\x9c', '\x9d', '\x9e', '\x9f', '\xa0', '\xa1', '\xa2', '\xa3', '\xa4', '\xa5', '\xa6', '\xa7', '\xa8', '\xa9', '\xaa', '\xab', '\xac', '\xad', '\xae', '\xaf', '\xb0', '\xb1', '\xb2', '\xb3', '\xb4', '\xb5', '\xb6', '\xb7', '\xb8', '\xb9', '\xba', '\xbb', '\xbc', '\xbd', '\xbe', '\xbf', '\xc0', '\xc1', '\xc2', '\xc3', '\xc4', '\xc5', '\xc6', '\xc7', '\xc8', '\xc9', '\xca', '\xcb', '\xcc', '\xcd', '\xce', '\xcf', '\xd0', '\xd1', '\xd2', '\xd3', '\xd4', '\xd5', '\xd6', '\xd7', '\xd8', '\xd9', '\xda', '\xdb', '\xdc', '\xdd', '\xde', '\xdf', '\xe0', '\xe1', '\xe2', '\xe3', '\xe4', '\xe5', '\xe6', '\xe7', '\xe8', '\xe9', '\xea', '\xeb', '\xec', '\xed', '\xee', '\xef', '\xf0', '\xf1', '\xf2', '\xf3', '\xf4', '\xf5', '\xf6', '\xf7', '\xf8', '\xf9', '\xfa', '\xfb', '\xfc', '\xfd', '\xfe', '\xff');

function getChar(code) {
  return CHARTABLE[parseInt(code)];
}

function urldecode(url) {
  var res = "";

  var l = url.split("%");

  var res = l[0];

  for (var i = 1; i < l.length; ++i) {
    var s = l[i];
	    
    var code = parseInt('0x' + s.substr(0, 2));
	    
    res += getChar(code);
	    
    res += s.substr(2);
  }

  return res;
}


var gThread;
  
function getThread(runnable) {
  var nsIThread = Components.interfaces.nsIThread;

  if (runnable) {
      
      if (FF3) {
	  if (gThread) { // FF3 threads can be reused
	      
	      gThread.dispatch(runnable, gThread.DISPATCH_NORMAL);
	      
	      return;
	  }
      }

      gThread = makeThread(runnable);
  }
}


/**
*  
*  UTF-8 data encode / decode
*  http://www.webtoolkit.info/
*
**/

var Utf8 = {

    // public method for url encoding
    encode : function (string) {
        string = string.replace(/\r\n/g,"\n");
        var utftext = "";

        for (var n = 0; n < string.length; n++) {

            var c = string.charCodeAt(n);

            if (c < 128) {
                utftext += String.fromCharCode(c);
            }
            else if((c > 127) && (c < 2048)) {
                utftext += String.fromCharCode((c >> 6) | 192);
                utftext += String.fromCharCode((c & 63) | 128);
            }
            else {
                utftext += String.fromCharCode((c >> 12) | 224);
                utftext += String.fromCharCode(((c >> 6) & 63) | 128);
                utftext += String.fromCharCode((c & 63) | 128);
            }

        }

        return utftext;
    },

    // public method for url decoding
    decode : function (utftext) {
        var string = "";
        var i = 0;
        var c = c1 = c2 = 0;

        while ( i < utftext.length ) {

            c = utftext.charCodeAt(i);

            if (c < 128) {
                string += String.fromCharCode(c);
                i++;
            }
            else if((c > 191) && (c < 224)) {
                c2 = utftext.charCodeAt(i+1);
                string += String.fromCharCode(((c & 31) << 6) | (c2 & 63));
                i += 2;
            }
            else {
                c2 = utftext.charCodeAt(i+1);
                c3 = utftext.charCodeAt(i+2);
                string += String.fromCharCode(((c & 15) << 12) | ((c2 & 63) << 6) | (c3 & 63));
                i += 3;
            }

        }

        return string;
    }

}


function getData(wikiData, key, type, uri, query) {
	
    trace("Zipedia: getData()");

    try {

	wikiData.load();

	trace("Zipedia:new Wiki() >>");

	var wiki = new Wiki(wikiData);

	trace("Zipedia:new Wiki() <<");

	var data = "hello sailor";

	var contentType = "text/html";

	switch (type) {
	case "wikitools":
	    var path = key;

	    var r;

	    if (path == "search") {
		data = wiki.handleSearch("", query);
	    } else if (null != (r = path.match("example/(.*)"))) {
		data = wiki.handleExample(r[1]);
	    } else {
		data = NOT_FOUND_ERROR.replace(/%URL%/g, uri.spec);
	    }
	  
	    break;
	case "raw":
	    data = wiki.handleRaw(key);
	    contentType = "text/plain";
	    break;
	case "wiki":

	    trace("Zipedia: wiki: key=" + key);

	    switch (key) {
	    case "":
		data = wiki.redirect("wiki", "MainPage");
		break;
	    default:
		data = wiki.handleWiki(key, uri);
		break;
	    }
	    break;
	default:
	    data = NOT_FOUND_ERROR.replace(/%URL%/g, uri.spec);
	    break;
	}

	return [data, contentType];

    } finally {
	wikiData.close();
    }
}

function makeWikiDataInstance()   {

    /* 
     * Important: don't instanciate XPCOM data access object inside runnable.run()
     * do it here instead on main thread.
     */

    var wikiData;

    var backend =  YAGI_getPref("zipedia.backend.", "data");
	    
    trace("Zipedia: Wiki() implementation: backend=" + backend);
	    
    switch (backend) {
    case "sqlite":
	trace("Zipedia: Wiki() createInstance(Components.interfaces.zpwkIWikiDataSqlite) >>");
	wikiData = Components.classes["@nargila.org/wiki-data/sqlite;1"].createInstance(Components.interfaces.zpwkIWikiDataSqlite);
	trace("Zipedia: Wiki() createInstance(Components.interfaces.zpwkIWikiDataSqlite) <<");
	break;
    case "java":
	wikiData = Components.classes["@nargila.org/wiki-data/java;1"].createInstance(Components.interfaces.zpwkIWikiDataJava);
	break;
    default:
	throw "HDIGH!";
    }

    return wikiData;
}





/*
  Shameless copy from Torisugari's (Kyoto, Japan) posting on 
  http://forums.mozillazine.org/viewtopic.php?p=921150#921150
*/
function getContents(aURL){ 
  var ioService=Components.classes["@mozilla.org/network/io-service;1"] 
    .getService(Components.interfaces.nsIIOService); 
  var scriptableStream=Components .classes["@mozilla.org/scriptableinputstream;1"] 
    .getService(Components.interfaces.nsIScriptableInputStream); 
  var channel=ioService.newChannel(aURL,null,null); 
  var input=channel.open(); 
  scriptableStream.init(input); 
  var str=scriptableStream.read(input.available()); 
  scriptableStream.close(); 
  input.close(); return str; 
}

function _makePipe() {
  var pipe = Components.classes["@mozilla.org/pipe;1"].
    createInstance(Components.interfaces.nsIPipe);
  
  pipe.init(false, false, 0, 0, null);
  
  return pipe;
}


/**
 * load external javascript files: e.g. loadScript("chrome://yagitalk/contents/scripts/utilitites.js")
 */
function loadScript(url) {
  var sl = Components.classes["@mozilla.org/moz/jssubscript-loader;1"].
    createInstance(Components.interfaces.mozIJSSubScriptLoader);

  sl.loadSubScript(url);
}

loadScript("chrome://zipedia/content/scripts/utilities.js");
loadScript("chrome://zipedia/content/wiki/wiki.js");

